/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/

package org.jboss.jms.server.remoting;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.EOFException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.Socket;
import java.util.Map;

import org.jboss.jms.client.remoting.ClientSocketWrapper;
import org.jboss.logging.Logger;
import org.jboss.remoting.serialization.SerializationStreamFactory;
import org.jboss.remoting.serialization.SerializationManager;

/**
 * @author <a href="mailto:tom.elrod@jboss.com">Tom Elrod</a>
 *
 * $Id$
 */
public class ServerSocketWrapper extends ClientSocketWrapper
{
   // Constants ------------------------------------------------------------------------------------

   final static private Logger log = Logger.getLogger(ServerSocketWrapper.class);

   // Static ---------------------------------------------------------------------------------------

   // Attributes -----------------------------------------------------------------------------------

   private static boolean trace = log.isTraceEnabled();

   // Constructors ---------------------------------------------------------------------------------

   public ServerSocketWrapper(Socket socket) throws Exception
   {
      super(socket);

      log.debug(this + " constructed");
   }

   public ServerSocketWrapper(Socket socket, Map metadata, Integer timeout) throws Exception
   {
      super(socket, metadata, timeout);

      log.debug(this + " constructed using a timeout of " + timeout + " ms");
   }

   // Public ---------------------------------------------------------------------------------------

   public void checkConnection() throws IOException
   {
      // Perform acknowledgement to convince client that the socket is still active
      byte ACK = 0;

      try
      {
         ACK = ((ObjectInputStream)getInputStream()).readByte();
      }
      catch(EOFException eof)
      {
         if(trace)
         {
            log.trace("socket timeout is set to: " + getTimeout());
            log.trace("EOFException waiting on ACK in readByte().");
         }
         throw eof;
      }
      catch(IOException e)
      {
         log.trace("IOException when reading in ACK", e);
         throw e;
      }

      if(trace) { log.trace("acknowledge read byte " + Thread.currentThread()); }

      ObjectOutputStream out = (ObjectOutputStream)getOutputStream();
      out.writeByte(ACK);
      out.flush();
      out.reset();
   }

   public String toString()
   {
      return "ServerSocketWrapper[" + getSocket() + "]";
   }

   // Package protected ----------------------------------------------------------------------------

   // Protected ------------------------------------------------------------------------------------

   protected ObjectInputStream createInputStream(String serializationType, Socket socket)
      throws IOException
   {
      if (trace) { log.trace(this + " creating input stream"); }

      BufferedInputStream bin = new BufferedInputStream(socket.getInputStream());
      SerializationManager sm = SerializationStreamFactory.getManagerInstance(serializationType);

      if (trace) { log.trace(this + " creating input stream using " + sm); }

      ObjectInputStream ois = sm.createInput(bin, null);

      if (trace) { log.trace(this + " created and returned " + ois); }

      return ois;
   }

   protected ObjectOutputStream createOutputStream(String serializationType, Socket socket)
         throws IOException
   {
      if (trace) { log.trace(this + " creating output stream"); }

      BufferedOutputStream bout = new BufferedOutputStream(socket.getOutputStream());
      SerializationManager sm = SerializationStreamFactory.getManagerInstance(serializationType);

      if (trace) { log.trace(this + " creating output stream using " + sm); }

      ObjectOutputStream oos = sm.createOutput(bout);

      oos.flush();

      if (trace) { log.trace(this + " created and returned " + oos); }

      return oos;
   }

   // Private --------------------------------------------------------------------------------------

   // Inner classes --------------------------------------------------------------------------------

}
